
# THIS FILE IS GENERATED FROM OASYS SETUP.PY
short_version = '1.2.68'
version = '1.2.68'
full_version = '1.2.68'
git_revision = '5cd7a7dc9f460ae690ab30f5c73afbbcad747eda'
release = True

if not release:
    version = full_version
    short_version += ".dev"
